@@ -7,12 +7,12 @@ import oapackage as oa  # type: ignore
 import pandas as pd  # type: ignore
 
 from .main import basic_factor_matrix, custom_design, twlp
-from .relation import Relation, num2gen
+from .relation import Relation, num2gen, gen2num
 
 
 class Design:
     """
-    Regular design with four- and two-level factors.
+    Regular design with four-level and two-level factors.
 
     Parameters
     ----------
@@ -30,22 +30,28 @@ class Design:
         Number of runs
     m : int
         Number of four-level factors
-    cols: List[int]
-        List of the column numbers of all the two-level factors of the design.
+    n : int
+        Number of two-level factors in the design
     k : int
         Number of basic factors. Equals to the log2 of the runsize.
-    pf : List[List[int]]
-        List of the pseudo-factors triplets (as lists of integers), used to
-        define the four-level factors.
-    bf: List[int]
-        List of the column numbers of the basic factors not used
-        in the pseudo-factors `pf`.
     p : int
         Number of added factors
+
+    bf : List[int]
+        List of the column numbers of the basic factors not used in the pseudo-factors
+        `pf`.
+    pf : List[List[int]]
+        List of the pseudo-factors triplets (as lists of integers), used to define the
+        four-level factors.
     af : List[int]
         List of the column numbers of the added factors
-    n: int
-        Number of two-level factors in the design
+    cols : List[int]
+        List of the column numbers of all the two-level factors of the design.
+
+    array : np.ndarray
+        Design matrix of the design with the four-level factors first and then the
+        two-level factors. The two-level factors are ordered by column number and not
+        kept in the original order.
     """
 
     # TODO: refactor the docstring of the Design class
@@ -192,10 +198,18 @@ class Design:
         else:
             return wlp_list[0 : (max_length - 2)]
 
+    def resolution(self) -> Optional[int]:
+        """
+        Compute the resolution of the design.
+        """
+        wlp = self.wlp()
+        res = next((i + 3 for i, x in enumerate(wlp) if x), None)
+        return res
+
     def flatten(self, zero_coding: bool = True) -> np.ndarray:
         """
-        Flatten each four-level factor of the design into two independent two-level
-        factors.
+        Flatten each four-level factor of the design into three two-level
+        pseudo-factors.
 
         Parameters
         ----------
@@ -224,7 +238,7 @@ class Design:
 
     def tfi_clearance(self) -> Dict[str, Dict[str, object]]:
         """
-        Compute clearance of all two-factor interactions in the design.
+        Compute clarity of all two-factor interactions in the design.
 
         There are three types of two-factor interactions in a design:
 
@@ -239,7 +253,7 @@ class Design:
         interactions, and '2-2 clear' when it is not aliased with any 2-2 interactions.
         When an interaction is 4-4 clear, 4-2 clear, and 2-2 clear, we call it a
         "totally clear (TC)" interaction.
-        Defining how clear is an interaction, is called the *clearance* of an
+        Defining how clear is an interaction, is called the *clarity* of an
         interaction.
 
         Returns
@@ -249,20 +263,28 @@ class Design:
         where the key is the name of the interaction, and the value is another
         dictionary containing four entries:
 
+            - 'clear': the interaction is clear from main effects
             - '4-4': the interaction is '4-4' clear (True or False)
             - '4-2': the interaction is '4-2' clear (True or False)
             - '2-2': the interaction is '2-2' clear (True or False)
             - 'Type': the type of the interaction, either '4-4', '4-2', or '2-2'
 
         """
-        # 4lvl PF are labeled as uppercase + number
+        # LABELS
+        ### MAIN EFFECTS
+
+        # 4lvl PF are labeled as uppercase + index i = 1,2,3
         label_list_4lvl_pf = [
             [f"{chr(65 + i)}{j}" for j in [1, 2, 3]] for i in range(self.m)
         ]
         # 2lvl factors labeled as lowercase
         label_list_2lvl = [chr(97 + i) for i in range(self.n)]
+
+        ### TWO-FACTOR-INTERACTIONS
+
         # Combinations of PF from the same four-level factors cannot be considered as
-        # for any PF, p1 x p2 = p3
+        # for any PF, p1 x p2 = p3 so combine BETWEEN the pseudo-factors triplets and
+        # not within
         if self.m > 1:
             label_list_44_tfi = list(
                 chain(
@@ -272,31 +294,40 @@ class Design:
                     ]
                 )
             )
+        # There are no 4-4 interactions for m=1
         else:
             label_list_44_tfi = []
+        # A 4-2 interaction is any combination of a PF and a two-level factor
         label_list_42_tfi = [
             f"{i}.{j}"
             for i in list(chain(*label_list_4lvl_pf))
             for j in label_list_2lvl
         ]
         label_list_22_tfi = [f"{i}.{j}" for i, j in combinations(label_list_2lvl, 2)]
+
         # All combinations of labels are all the TFI
         label_list_tfi = label_list_44_tfi + label_list_42_tfi + label_list_22_tfi
-        # tfi can be one of three types: 4-4, 4-2, or 2-2
-        type_list_tfi = (
-            ["4-4"] * len(label_list_44_tfi)
-            + ["4-2"] * len(label_list_42_tfi)
-            + ["2-2"] * len(label_list_22_tfi)
-        )
-        # All alias-related variables are NA and will be filled later on
-        tfi = dict()
-        for tfi_idx, label in enumerate(label_list_tfi):
-            tfi_type = type_list_tfi[tfi_idx]
-            tfi[label] = {"4-4": True, "4-2": True, "2-2": True, "Type": tfi_type}
-        # Zero coding is needed to perform multiplication of factors
+
+        # MATRIX
+
+        # Zero coding (0/1) is needed to perform multiplication of factors
         flat_array_coded = self.flatten(zero_coding=False)
-        four_level_pf_range = range(3 * self.m)
+
+        ### MAIN EFFECTS
+
         two_level_fac_range = range(3 * self.m, (3 * self.m + self.n))
+        matrix_2_me = flat_array_coded[:, two_level_fac_range]
+        four_level_pf_range = range(3 * self.m)
+        matrix_4_me = flat_array_coded[:, four_level_pf_range]
+
+        ### TWO-FACTOR INTERACTIONS
+
+        matrix_22_tfi = np.hstack(
+            [
+                flat_array_coded[:, [i]] * flat_array_coded[:, [j]]
+                for i, j in combinations(two_level_fac_range, 2)
+            ]
+        )
         matrix_42_tfi = np.hstack(
             [
                 flat_array_coded[:, [i]] * flat_array_coded[:, [j]]
@@ -304,12 +335,8 @@ class Design:
                 for j in two_level_fac_range
             ]
         )
-        matrix_22_tfi = np.hstack(
-            [
-                flat_array_coded[:, [i]] * flat_array_coded[:, [j]]
-                for i, j in combinations(two_level_fac_range, 2)
-            ]
-        )
+
+        # We only compute 44 interactions if m > 1
         if self.m > 1:
             list_44_tfi = [
                 [
@@ -320,12 +347,50 @@ class Design:
                 for x, y in combinations(range(self.m), 2)
             ]
             matrix_44_tfi = np.hstack(list(chain(*list_44_tfi)))
+            effect_mat = np.concatenate(
+                (matrix_44_tfi, matrix_42_tfi, matrix_22_tfi, matrix_2_me, matrix_4_me),
+                axis=1,
+            )
+            # We don't need to define clarity for the ME
             tfi_mat = np.concatenate(
-                (matrix_44_tfi, matrix_42_tfi, matrix_22_tfi), axis=1
+                (matrix_44_tfi, matrix_42_tfi, matrix_22_tfi),
+                axis=1,
             )
         else:
-            tfi_mat = np.concatenate((matrix_42_tfi, matrix_22_tfi), axis=1)
-        tfi_aliasing_mat = tfi_mat.T @ tfi_mat
+            effect_mat = np.concatenate(
+                (matrix_42_tfi, matrix_22_tfi, matrix_2_me, matrix_4_me),
+                axis=1,
+            )
+            # We don't need to define clarity for the ME
+            tfi_mat = np.concatenate(
+                (matrix_42_tfi, matrix_22_tfi),
+                axis=1,
+            )
+
+        # ALIASING
+
+        # tfi can be one of three types: 4-4, 4-2, or 2-2
+        type_list_tfi = (
+            ["4-4"] * len(label_list_44_tfi)
+            + ["4-2"] * len(label_list_42_tfi)
+            + ["2-2"] * len(label_list_22_tfi)
+            + ["clear"] * (len(label_list_2lvl) + 3 * len(label_list_4lvl_pf))
+        )
+
+        # All alias-related variables are NA and will be filled later on
+        tfi = dict()
+        for tfi_idx, label in enumerate(label_list_tfi):
+            tfi_type = type_list_tfi[tfi_idx]
+            tfi[label] = {
+                "clear": True,
+                "4-4": True,
+                "4-2": True,
+                "2-2": True,
+                "Type": tfi_type,
+            }
+
+        # Rows are the TFI we are investigating
+        tfi_aliasing_mat = tfi_mat.T @ effect_mat
         # All TFI are aliased with themselves
         np.fill_diagonal(tfi_aliasing_mat, 0)
         row, col = np.nonzero(tfi_aliasing_mat)
@@ -367,12 +432,12 @@ class Design:
             for clear_type in ["4-4", "4-2", "2-2"]:
                 if interaction[clear_type]:
                     column_label = f"{clear_type} clear"
-                    clarity_df.loc[int_type, column_label] += 1
-                    clarity_df.loc["Any type", column_label] += 1
+                    clarity_df.loc[int_type, column_label] += 1  # type: ignore
+                    clarity_df.loc["Any type", column_label] += 1  # type: ignore
                     all_clear += 1
                 if all_clear == 3:
-                    clarity_df.loc[int_type, "Totally clear"] += 1
-                    clarity_df.loc["Any type", "Totally clear"] += 1
+                    clarity_df.loc[int_type, "Totally clear"] += 1  # type: ignore
+                    clarity_df.loc["Any type", "Totally clear"] += 1  # type: ignore
         return clarity_df
 
     def clear(self, interaction_type: str, clear_from: str) -> int:
@@ -490,3 +555,34 @@ class Design:
         new_cols = [i for i in self.af]
         new_cols.remove(number)
         return Design(self.runsize, self.m, new_cols)
+
+
+def from_array(mat: np.ndarray, zero_coded: bool = True):
+    # Define runsize
+    run_size, n_fac = mat.shape
+    # Define number of four-level factors
+    levels = [len(np.unique(mat[:, i])) for i in range(n_fac)]
+    n_flvl = sum([i == 4 for i in levels])
+    # Sort matrix and extract two-level part
+    sorting_list = [mat[:, i] for i in range(n_flvl)]
+    sorted_mat = mat[np.lexsort(sorting_list)]
+    tlvl_mat = sorted_mat[:, [i == 2 for i in levels]]
+    if not zero_coded:
+        tlvl_mat = (tlvl_mat + 1) / 2
+    # Compute all possible columns for that runsize
+    k = int(np.log2(run_size))
+    bf_matrix = basic_factor_matrix(k=k, zero_coding=True)
+    full_generator_matrix = np.zeros((k, 2**k - 1), dtype=int)
+    labels_list = []
+    for i in range(2**k - 1):
+        bin_string = f"{(i + 1):0{k}b}"[::-1]
+        full_generator_matrix[:, i] = [int(j) for j in bin_string]
+        gen = [chr(i + 97) for i, x in enumerate(bin_string) if int(x) == 1]
+        labels_list.append("".join(gen))
+    full_design_matrix = np.matmul(bf_matrix, full_generator_matrix) % 2
+    cols_mat = np.matmul(full_design_matrix.T * 2 - 1, tlvl_mat * 2 - 1)
+    cols_id, _ = np.nonzero(cols_mat)
+    cols_gen = [labels_list[i] for i in cols_id]
+    cols_num = [gen2num(i) for i in cols_gen]
+    added_cols = [i for i in cols_num if np.log2(i) % 1 != 0]
+    return Design(run_size, n_flvl, added_cols)
